#include <assert.h>
#include <stdio.h>
#include <string.h>

#include "object.h"
#include "Attribute.h"
#include "AttributeCache.h"
#include "LandscapeObject.h"
#include "LandscapeObject3D.h"
#include "EntityInstance.h"
#include "RelationInstance.h"
#include "LandscapeClassObject.h"
#include "EntityClass.h"
#include "RelationClass.h"
#include "EntityCache.h"
#include "TAStreamTokenizer.h"
#include "LandscapeTokenStream.h"
#include "TAInput.h"

/* This layer of software is merely concerned with representing the TA graph structures 
   Any access to update methods at this level will perform no signalling of changes
   These methods should only be used during the actual loading of an initial TA graph
 */

CEntityClass *
CTAInput::newEntityClass(const char *idP)
{
	CEntityClass	*entityClassP;

	entityClassP          = new CEntityClass(idP);
	entityClassP->m_nextP = m_entityClassListP;
	m_entityClassListP    = entityClassP;
	return entityClassP;
}

CEntityClass *
CTAInput::getEntityClass(const char *idP)
{
	CEntityClass *entityClassP;

	entityClassP = m_entityClassListP;
	if (idP) {
		for (; entityClassP; entityClassP = entityClassP->m_nextP) {
			if (entityClassP->hasId(idP)) {
				break;
	}	}	}
	return (entityClassP); 
}
	
CEntityClass *
CTAInput::addEntityClass(const char *idP) 
{
	CEntityClass *entityClassP = getEntityClass(idP); 
	if (!entityClassP) {
		entityClassP = newEntityClass(idP);
	}
	return entityClassP;
}
	
CRelationClass *
CTAInput::newRelationClass(const char *idP)
{
	CRelationClass	*relationClassP;

	relationClassP          = new CRelationClass(idP);
	relationClassP->m_nextP = m_relationClassListP;
	m_relationClassListP    = relationClassP;
	return relationClassP;
}

CRelationClass *
CTAInput::getRelationClass(const char *idP)
{
	CRelationClass *relationClassP;

	for (relationClassP = m_relationClassListP; relationClassP; relationClassP = relationClassP->m_nextP) {
		if (relationClassP->hasId(idP)) {
			break;
	}	}
	return(relationClassP); 
}

CRelationClass *
CTAInput::addRelationClass(const char *idP) 
{
	CRelationClass *rcP = getRelationClass(idP);		// Lookup in the hash table

	if (!rcP) {
		rcP = newRelationClass(idP);
	}
	return rcP;
}

CTAInput::CTAInput(void) 
{
	m_rootInstanceP      = 0;
	m_entityClassListP   = 0;
	m_relationClassListP = 0;
	m_entityBaseClassP   = newEntityClass("$ENTITY");		// $ENTITY
	m_relationBaseClassP = newRelationClass("$RELATION");	// $RELATION
	m_entityCacheP       = new CEntityCache();
}

CEntityClass *
CTAInput::getEntityBaseClass(void) const
{
	return m_entityBaseClassP;
}

CRelationClass *
CTAInput::getRelationBaseClass(void) const
{
	return m_relationBaseClassP;
}

CEntityInstance *
CTAInput::getRootInstance(void) const
{
	return m_rootInstanceP;
}

CEntityInstance *
CTAInput::getCache(const char *idP) const
{
	return m_entityCacheP->get(idP);
} 

void 
CTAInput::putCache(CEntityInstance *eP) const
{
	m_entityCacheP->put(eP);
}

// --------------
// Update methods
// --------------

void 
CTAInput::createRootInstance(void)
{
	m_rootInstanceP = newCachedEntity(m_entityBaseClassP, "$ROOT"); 	// $ROOT
}

void 
CTAInput::setupUniversalScheme() 
{
	m_relationBaseClassP->addRelationConstraint(m_entityBaseClassP, m_entityBaseClassP);
}

CEntityInstance *
CTAInput::newCachedEntity(CEntityClass *ecP, const char *idP)
{
	assert(ecP);

	CEntityInstance *eP = ecP->newEntity(idP); 
	putCache(eP);
	return eP;
}

// ------------------
// TA Reading methods
// ------------------

void 
CTAInput::processSchemeTuples(CLandscapeTokenStream *tsP) 
{
	CRelationClass	*rc1P, *rc2P;
	const char		*verbP, *objectP;
	CEntityClass	*ec1P, *ec2P;

	while (tsP->nextSchemaTriple()) {

		verbP    = tsP->m_verbP;
		objectP  = tsP->m_objectP;

		if (!strcmp(verbP, "$INHERIT")) {							// $INHERIT
			switch (tsP->m_relations) {
			case 0:	/* Neither term is a relation */
				if (!strcmp(objectP, "$ENTITY")) {
					fprintf(stderr, "*** Error (%d):%s %s %s: Improper use of $ENTITY with $INHERIT\n", tsP->lineno(), verbP, objectP, tsP->m_subjectP);
					break;
				}
				ec1P = addEntityClass(objectP); 
				ec2P = addEntityClass(tsP->m_subjectP);
				ec1P->setParentClass(ec2P);
				break;
			case 1:	/* First  is relation		*/
			case 2:	/* Second is relation		*/
				fprintf(stderr, "*** Error (%d):Mismatched entity/relation with $INHERIT -- presuming both relations\n", tsP->lineno());
			case 3:	/* Both term is a relation	*/
				if (!strcmp(objectP, "$ENTITY")) {
					fprintf(stderr, "*** Error (%d):%s %s %s: Improper use of $RELATION with $INHERIT\n", tsP->lineno(), verbP, objectP, tsP->m_subjectP);
					break;
				}
				rc1P = addRelationClass(objectP);
				rc2P = addRelationClass(tsP->m_subjectP);
				rc1P->setParentClass(rc2P);
				break;
			}

		} else { 
			if (tsP->m_relations != 0) {
				fprintf(stderr, "*** Error (%d):Cannot create relationships between relationships\n", tsP->lineno());
			} else {
				ec1P = addEntityClass(objectP); 
				ec2P = addEntityClass(tsP->m_subjectP);
				rc1P = addRelationClass(verbP);
				rc1P->addRelationConstraint(ec1P, ec2P);	// If not already present
		}	}
	}
}

void
CTAInput::processFactTuples(CLandscapeTokenStream *tsP)
{
	const char		*verbP, *objectP, *subjectP;
	CEntityInstance	*eP, *srcP, *dstP;
	CEntityClass	*ecP, *ec1P;
	CRelationClass	*rcP;

	while (tsP->nextFactTriple()) {

		verbP    = tsP->m_verbP;
		objectP  = tsP->m_objectP;
		subjectP = tsP->m_subjectP;

		if (!strcmp(verbP, "$INSTANCE")) {

			// 
			// $INSTANCE instanceId classId
			//

			if (!strcmp(objectP, "$ROOT")) {
				eP = getRootInstance();
			} else {
				eP = getCache(objectP);
			}

			ecP = addEntityClass(subjectP);
			if (!eP) {
				// New $INSTANCE
				eP = newCachedEntity(ecP, objectP);
			} else {
				ec1P = eP->getEntityClass();
				if (ecP != ec1P) {
					if (ec1P /* Special case for $ROOT */  && ec1P != m_entityBaseClassP) {
						fprintf(stderr, ">>> Warning (%d):Redeclaration of %s from instance of %s to instance of %s", tsP->lineno(), eP->getId(), ec1P->getId(), ecP->getId());
					}
					eP->setParentClass(ecP);
			}	}
		} else {

			// Relation tuple definition
			//
			// relationClass entityInstance entityInstance
			//

			srcP = getCache(objectP);
			if (!srcP) {
				srcP = newCachedEntity(m_entityBaseClassP, objectP);
			}
			dstP = getCache(subjectP);
			if (!dstP) {
				dstP = newCachedEntity(m_entityBaseClassP, subjectP);
			}
			rcP  = addRelationClass(verbP);

			new CRelationInstance(rcP, srcP, dstP);
		}
	}
}

void 
CTAInput::attachBaseClasses()
{
	CRelationClass		*rc1P;
	CEntityClass		*ec1P;
	CEntityInstance		*rootP;
	
	rootP = getRootInstance();
	if (!rootP->getParentClass()) {
		rootP->setParentClass(m_entityBaseClassP);
	}
	for (ec1P = m_entityClassListP; ec1P; ec1P = ec1P->m_nextP) {
		if (ec1P != m_entityBaseClassP  && !ec1P->getParentClass()) {
			ec1P->setParentClass(m_entityBaseClassP);
	}	}
	
	for (rc1P = m_relationClassListP; rc1P; rc1P = rc1P->m_nextP) {
		if (rc1P != m_relationBaseClassP && !rc1P->getParentClass()) {
			rc1P->setParentClass(m_relationBaseClassP);
}	}	}
	
void 
CTAInput::parseStream(const char *filenameP) 
{
	int	schemeSetup  = false;

	CLandscapeTokenStream *tsP = new CLandscapeTokenStream(this);

	tsP->open(filenameP);
	for (;;) {
		int sec = tsP->nextSection();

		if (sec == EOFTOKEN) {
			break;
		}
		switch(sec) {

		case SCHEME_TUPLE:

			schemeSetup = true;
			processSchemeTuples(tsP);
			break;

		case SCHEME_ATTRIBUTE:

			if (!schemeSetup) {
				schemeSetup = 1;
				setupUniversalScheme();
			}
			tsP->processSchemeAttributes();
			break;

		case FACT_TUPLE:

			if (!schemeSetup) {
				schemeSetup = 1;
				setupUniversalScheme();
			}

			processFactTuples(tsP);
			break;

		case FACT_ATTRIBUTE:

			tsP->processFactAttributes();
			break;
		}
	}
	tsP->close();
	attachBaseClasses();

} // end parseStream() 


void
CTAInput::loadTA(const char *filenameP) 
{
	createRootInstance();
	parseStream(filenameP);
	putCache(m_rootInstanceP);
}


